from __future__ import annotations

from dataclasses import dataclass

from .path import Path


@dataclass
class Config:
    hostname: str = "https://music.com"
    backend_port: int = 13000

    @classmethod
    def load(cls) -> Config:
        return cls(**Path.config.yaml)


config = Config.load()



import tbhandler
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse

from ..client.response_types import ArtistInfo, Track
from . import artistmanager
from .artistmanager import DisplayArtist, DisplaySong

app = FastAPI()

origins = [f"http://localhost:{8080}"]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=origins,
    allow_headers=origins,
)


@app.exception_handler(Exception)  # type: ignore
async def handler(_: Request, exc: Exception) -> PlainTextResponse:
    tbhandler.show(exit_after=False, repeat=False)
    return PlainTextResponse(str(exc), status_code=400)


@app.get("/newartist")
async def get(name: str) -> list[DisplayArtist]:
    return artistmanager.search_artists(name)


@app.get("/newsong")  # type: ignore[no-redef]
async def get(name: str) -> list[DisplaySong]:
    return artistmanager.search_song(name)


@app.get("/addartist")  # type: ignore[no-redef]
async def get(id: str, name: str) -> None:
    return artistmanager.add_artist(id, name)


@app.get("/addsong")  # type: ignore[no-redef]
async def get(id: str) -> None:
    return artistmanager.add_song(id)


@app.get("/changeartist")  # type: ignore[no-redef]
async def get(id: str) -> None:
    return artistmanager.change_artist(id)


@app.get("/recommendedartists")  # type: ignore[no-redef]
async def get() -> list[ArtistInfo]:
    return artistmanager.recommendations()


@app.get("/recommendedsongs")  # type: ignore[no-redef]
async def get() -> list[Track]:
    return artistmanager.song_recommendations()


@app.get("/artists")  # type: ignore[no-redef]
async def get() -> list[dict[str, str]]:
    return artistmanager.artists()



from ..utils import Path
from . import downloader, jobs, uploader


def start_download() -> None:
    songs = jobs.get()
    if songs:
        downloader.Downloader(songs).start()
    if not Path.processed_songs.is_empty():
        uploader.start()



from __future__ import annotations

import random
from dataclasses import dataclass

from ..artist import Artist, Artists
from ..artist.artists import artists as artists_collection
from ..client import response_types, spotapi
from ..client.response_types import ArtistInfo, Track
from ..downloads import jobs
from ..utils import Path


@dataclass
class DisplaySong(response_types.Track):
    downloaded: bool = False

    @classmethod
    def from_track(cls, track: response_types.Track) -> DisplaySong:
        # don't use dict() method because it serializes all attributes recursively
        attributes = track.__dict__
        attributes["downloaded"] = jobs.is_downloaded(track)
        return DisplaySong(**attributes)


@dataclass
class DisplayArtist(response_types.ArtistInfo):
    added: bool = False

    @classmethod
    def from_artist_info(cls, artist: response_types.ArtistInfo) -> DisplayArtist:
        attributes = artist.__dict__
        attributes["added"] = artist.id in artists_collection.artists
        return DisplayArtist(**attributes)


def artists() -> list[dict[str, str]]:
    loaded_artists = Artists()
    api_infos = spotapi.artists(loaded_artists.ids)
    loaded_artists_list = loaded_artists.artist_list()
    infos = [info.dict() | a.dict() for a, info in zip(loaded_artists_list, api_infos)]
    return infos


def add_artist(id: str, name: str) -> None:
    loaded_artists = Artists()
    artist = Artist(id, name)
    loaded_artists.add(artist)
    loaded_artists.save()


def add_song(id: str) -> None:
    tracks = spotapi.songs([id])
    jobs.add(tracks)


def search_artists(name: str) -> list[DisplayArtist]:
    return [
        DisplayArtist.from_artist_info(artist) for artist in spotapi.search_artist(name)
    ]


def search_song(name: str) -> list[DisplaySong]:
    return [DisplaySong.from_track(song) for song in spotapi.search_song(name)]
    # TODO: add youtube songs if spotify result list is empty


def change_artist(id: str) -> None:
    loaded_artists = Artists()
    loaded_artists[id].toggle_type()
    loaded_artists.save()


def recommendations(amount: int = 50, max_tries: int = 10) -> list[ArtistInfo]:
    ids = Artists().ids
    random.shuffle(ids)
    seed_ids = ids[:max_tries]

    freqs = Path.recommendations.yaml
    artist_recommendations: set[ArtistInfo] = set({})

    while len(artist_recommendations) < amount and seed_ids:
        id = seed_ids.pop(0)
        recommended_artists = [
            a for a in spotapi.related_artists(id) if a.id not in ids
        ]
        artist_recommendations.update(recommended_artists)
        for artist in recommended_artists:
            freqs[artist.id] = freqs.get(artist.id, 0) + 1

    Path.recommendations.yaml = freqs
    artist_recommendations_list = sorted(
        list(artist_recommendations), key=lambda r: freqs[r.id]
    )
    return artist_recommendations_list


def song_recommendations(seed_amount: int = 5) -> list[Track]:
    downloads = Path.download_ids.yaml
    ids = list(downloads.keys())
    random.shuffle(ids)
    seed_ids = ids[:seed_amount]

    names = {v for v in downloads.values()}
    songs = [
        s
        for s in spotapi.song_recommendations(seed_ids)
        if s.id not in downloads and jobs.full_name(s) not in names
    ]
    return songs



from collections.abc import Iterator
from typing import TypeVar

T = TypeVar("T")


def chunked(items: list[T], chunk_size: int = 50) -> Iterator[list[T]]:
    for i in range(0, len(items), chunk_size):
        yield items[i : i + chunk_size]



"""
Spotify changes the ids of its tracks sometimes run this file to give the downloaded
songs their most recent id.
"""

from typing import TypeVar

from ..client import spotapi
from ..utils import Path

T = TypeVar("T")
U = TypeVar("U")


def reverse(items: dict[T, U]) -> dict[U, T]:
    return {v: k for k, v in items.items()}


def main() -> None:
    downloads = Path.download_ids.yaml
    ids = list(downloads.keys())
    songs = spotapi.songs(ids)

    new_downloads = {song.id: title for song, title in zip(songs, downloads.values())}
    new_downloads = reverse(reverse(new_downloads))  # filter out duplicate song titles
    Path.download_ids.yaml = new_downloads


if __name__ == "__main__":
    main()



from collections.abc import Iterator

import pytest
from hypothesis import HealthCheck, given, settings, strategies
from hypothesis.strategies import SearchStrategy
from music.backend.utils import Path

NestedDict = dict[str, dict[str, str]]


def dictionary_strategy() -> SearchStrategy[dict[str, str]]:
    return strategies.dictionaries(keys=strategies.text(), values=strategies.text())


def nested_dictionary_generator() -> SearchStrategy[NestedDict]:
    return strategies.dictionaries(
        keys=strategies.text(),
        values=dictionary_strategy(),
    )


@pytest.fixture()  # type: ignore
def path() -> Iterator[Path]:
    with Path.tempfile() as path:
        yield path
    path.json_path.unlink(missing_ok=True)
    for removed_path in (path, path.json_path):
        assert not removed_path.exists()


@given(content=nested_dictionary_generator())  # type: ignore
@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,), max_examples=10)  # type: ignore
def test_yaml(path: Path, content: NestedDict) -> None:
    path.yaml = content
    assert path.yaml == content


@given(content=nested_dictionary_generator())  # type: ignore
@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,), max_examples=10)  # type: ignore
def test_json_usage(path: Path, content: NestedDict) -> None:
    path.json_path.json = content
    assert path.yaml == content


@given(content1=nested_dictionary_generator(), content2=nested_dictionary_generator())  # type: ignore
@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,), max_examples=10)  # type: ignore
def test_yaml_priority(path: Path, content1: NestedDict, content2: NestedDict) -> None:
    path.yaml = content1
    path.json_path.json = content2
    path.json_path.mtime = path.mtime - 1
    assert path.yaml == content1



from dataclasses import dataclass
from typing import Any, TypeVar

import dacite

T = TypeVar("T")


@dataclass
class Item:
    @classmethod
    def from_dict(cls: type[T], items: dict[str, Any]) -> T:
        return dacite.from_dict(cls, items, config=dacite.Config(strict=True))  # type: ignore



from dataclasses import dataclass

import cli
from spotdl.console import download as spotdl_downloader
from spotdl.download.downloader import Downloader as SpotifyDownloader

from ..client import get_spotdl_client
from ..utils import Path, iteration
from . import jobs, postprocessor

DOWNLOAD_RETRIES = 5


def clear_downloads() -> None:
    # existing downloads raise errors and can be removed because
    # successfully downloaded songs have already been moved to other folder
    Path.downloaded_songs.rmtree(remove_root=False)


@dataclass
class Downloader:
    songs: list[str]
    downloader: SpotifyDownloader = None
    chunk_size: int = 20

    def __post_init__(self) -> None:
        self.downloader = get_spotdl_client().downloader
        self.fix_song_count()

    def fix_song_count(self) -> None:
        num_songs = len(self.songs)
        self.downloader.progress_handler.set_song_count(num_songs)
        self.downloader.progress_handler.set_song_count = lambda count: None

    def start(self) -> None:
        for chunk in iteration.chunked(self.songs, chunk_size=self.chunk_size):
            self.download_songs_chunk(chunk)

    def download_songs_chunk(self, songs: list[str]) -> None:
        self.download_songs(songs)
        with cli.status("Processing downloads"):
            postprocessor.process_downloads()
        jobs.remove(songs)

    def download_songs(self, songs: list[str]) -> None:
        songs = [
            id_ if id_.startswith("http") else f"https://open.spotify.com/track/{id_}"
            for id_ in songs
        ]

        clear_downloads()
        tries_left = DOWNLOAD_RETRIES
        while songs and tries_left:
            spotdl_downloader.download(songs, self.downloader)
            songs = list(Path.downloaded_songs.glob("*.spotdlTrackingFile"))
            tries_left -= 1

        if songs and not tries_left:
            raise Exception("Max download retries reached")



from __future__ import annotations

import os
from dataclasses import dataclass, field

from .item import Item
from .path import Path


@dataclass
class SpotifyTokens(Item):
    client_id: str = field(default_factory=lambda: os.environ["SPOTIFY_CLIENT_ID"])
    client_secret: str = field(
        default_factory=lambda: os.environ["SPOTIFY_CLIENT_SECRET"]
    )


@dataclass
class Tokens(Item):
    genius: str = field(default_factory=lambda: os.environ["GENIUS_TOKEN"])
    spotify: SpotifyTokens = field(default_factory=lambda: SpotifyTokens())

    @classmethod
    def load(cls) -> Tokens:
        items = Path.tokens.json
        return cls.from_dict(items)


tokens = Tokens.load()



import cli
from music.backend.backend.main import Starter


def test_frontend_creation() -> None:
    def remove_frontend_dist() -> None:
        cli.run("rm -r", starter.frontend_dist, root=True)

    starter = Starter()
    existing = starter.frontend_dist.exists()
    if existing:
        remove_frontend_dist()
    starter.generate_frontend_distribution()
    if not existing:
        remove_frontend_dist()
